import torch
import torch.nn as nn

from .abstract_prior import Prior


class JointPrior(Prior):
    """Prior class that doesn't contribute to KL loss.

    Effectively a Dirac delta distribution given z.
    """

    def __init__(self, priors):
        dimensionality = sum(prior.dimensionality for prior in priors)
        super().__init__(dimensionality)

        if not isinstance(priors, nn.ModuleList):
            priors = nn.ModuleList(*priors)
        self.priors = priors

    def kl_divergence(self, z_params, reduction="sum"):
        kl = []

        start_ix = 0
        for prior in self.priors:
            end_ix = (start_ix + prior.dimensionality) - 1
            kl.append(
                prior(z_params[:, start_ix:end_ix], mode="kl", reduction=reduction)
            )
            start_ix = end_ix + 1

        kl = torch.cat(kl, axis=1)

        if reduction == "none":
            return kl
        elif reduction == "sum":
            return kl.sum(dim=-1)
        else:
            raise NotImplementedError(
                f"Reduction '{reduction}' not implemented for JointPrior"
            )

    def sample(self, z_params):
        samples = []

        start_ix = 0
        for prior in self.priors:
            end_ix = (start_ix + prior.dimensionality) - 1
            samples.append(prior(z_params[:, start_ix:end_ix], mode="sample"))
            start_ix = end_ix + 1

        return torch.cat(samples, axis=1)

    def forward(self, z_params, mode="kl", **kwargs):
        if mode == "kl":
            return self.kl_divergence(z_params)

        return self.sample(z_params)



from typing import Optional, Sequence
from omegaconf import DictConfig

import torch
import torch.nn as nn
from torch.nn.modules.loss import _Loss as Loss
from torch.optim.lr_scheduler import _LRScheduler as LRScheduler

from serotiny.models.base_model import BaseModel
from .priors import Prior, IsotropicGaussianPrior


def _latent_compose_function(z_parts, **kwargs):
    return z_parts


class BaseVAE(BaseModel):
    def __init__(
        self,
        encoder: nn.Module,
        decoder: nn.Module,
        latent_dim: int,
        x_label: str,
        beta: float = 1.0,
        id_label: Optional[str] = None,
        optimizer: torch.optim.Optimizer = torch.optim.Adam,
        lr_scheduler: Optional[LRScheduler] = None,
        reconstruction_loss: Loss = nn.MSELoss(reduction="none"),
        prior: Optional[Sequence[Prior]] = None,
        latent_compose_function=None,
        cache_outputs: Sequence = ("test",),
        **kwargs,
    ):
        """Instantiate a basic VAE model.
        Parameters
        ----------
        encoder: nn.Module
            Encoder network
        decoder: nn.Module
            Decoder network
        x_label: Optional[str] = None
        id_label: Optional[str] = None
        optimizer: torch.optim.Optimizer
            Optimizer to use
        beta: float = 1.0
            Beta parameter - the weight of the KLD term in the loss function
        reconstruction_loss: Loss
            Loss to be used for reconstruction. Can be a PyTorch loss or a class
            that respects the same interface,
            i.e. subclasses torch.nn.modules._Loss
        prior: Optional[Sequence[AbstractPrior]]
            List of prior specifications to use for latent space
        """
        super().__init__()

        self.reconstruction_loss = reconstruction_loss

        if not isinstance(encoder, (dict, DictConfig)):
            assert x_label is not None
            encoder = {x_label: encoder}
        self.encoder = nn.ModuleDict(encoder)

        if not isinstance(decoder, (dict, DictConfig)):
            assert x_label is not None
            decoder = {x_label: decoder}
        self.decoder = nn.ModuleDict(decoder)

        self.beta = beta
        self.latent_dim = latent_dim

        if prior is None:
            prior = IsotropicGaussianPrior()

        if not isinstance(prior, (dict, DictConfig)):
            assert x_label is not None
            prior = {x_label: prior}

        self.prior = nn.ModuleDict(prior)

        if latent_compose_function is None:
            latent_compose_function = _latent_compose_function
        self.latent_compose_function = latent_compose_function

    def calculate_rcl(self, x, x_hat, key):
        rcl_per_input_dimension = self.reconstruction_loss[key](x[key], x_hat[key])
        return rcl_per_input_dimension

    def calculate_elbo(self, x, x_hat, z):

        rcl_per_input_dimension = {}
        rcl_reduced = {}
        for key in x_hat.keys():
            rcl_per_input_dimension[key] = self.calculate_rcl(x, x_hat, key)
            if len(rcl_per_input_dimension[key].shape) > 0:
                rcl = (
                    rcl_per_input_dimension[key]
                    # flatten
                    .view(rcl_per_input_dimension[key].shape[0], -1)
                    # and sum across each batch element's dimensions
                    .sum(dim=1)
                )

                rcl_reduced[key] = rcl.mean()
            else:
                rcl_reduced[key] = rcl_per_input_dimension[key]

        kld_per_part = {
            part: self.prior[part](z_part, mode="kl", reduction="none")
            for part, z_part in z.items()
        }

        kld_per_part_summed = {
            part: kl.sum(dim=-1).mean() for part, kl in kld_per_part.items()
        }

        total_kld = sum(kld_per_part_summed.values())
        return (
            sum(rcl_reduced.values()) + self.beta * total_kld,
            rcl_reduced,
            total_kld,
            kld_per_part,
        )

    def sample_z(self, z_parts_params):
        return {
            part: prior(z_parts_params[part], mode="sample")
            for part, prior in self.prior.items()
        }

    def encode(self, batch):
        return {
            part: encoder(batch[part].float()) for part, encoder in self.encoder.items()
        }

    def decode(self, z_parts):
        z = self.latent_compose_function(z_parts)

        return (
            {part: decoder(z[part].float()) for part, decoder in self.decoder.items()},
            z,
        )

    def forward(self, batch, decode=False, compute_loss=False, **kwargs):

        z_parts_params = self.encode(batch)

        z_parts = self.sample_z(z_parts_params)

        x_hat, z_composed = self.decode(z_parts)

        if not decode:
            return z_parts_params, z_composed

        if not compute_loss:
            return x_hat, z_parts, z_parts_params, z_composed

        (
            loss,
            reconstruction_loss,
            kld_loss,
            kld_per_part,
        ) = self.calculate_elbo(batch, x_hat, z_parts_params)

        return (
            x_hat,
            z_parts,
            z_parts_params,
            z_composed,
            loss,
            reconstruction_loss,
            kld_loss,
            kld_per_part,
        )

    def log_metrics(self, stage, results, logger, batch_size):
        on_step = (stage == "val") | (stage == "train")

        for key, value in results.items():
            if (len(value.shape) == 0) | (len(value.shape) == 1):
                self.log(
                    f"{stage} {key}",
                    value,
                    logger=logger,
                    on_step=on_step,
                    on_epoch=True,
                    batch_size=batch_size,
                )

    def make_results_dict(
        self,
        stage,
        batch,
        loss,
        reconstruction_loss,
        kld_loss,
        kld_per_part,
        z_parts,
        z_parts_params,
        z_composed,
        x_hat,
    ):
        results = {
            "loss": loss,
            f"{stage}_loss": loss.detach().cpu(),  # for epoch end logging purposes
            "kld_loss": kld_loss.detach().cpu(),
        }

        for part, z_comp_part in z_composed.items():
            results.update(
                {
                    f"z_composed/{part}": z_comp_part.detach().cpu(),
                }
            )

        for part, recon_part in reconstruction_loss.items():
            results.update(
                {
                    f"reconstruction_loss/{part}": recon_part.detach().cpu(),
                }
            )

        for part, z_part in z_parts.items():
            results.update(
                {
                    f"z_parts/{part}": z_part.detach().cpu(),
                    f"z_parts_params/{part}": z_parts_params[part].detach().cpu(),
                    f"kld/{part}": kld_per_part[part].detach().float().cpu(),
                }
            )

        if self.hparams.id_label is not None:
            if self.hparams.id_label in batch:
                ids = batch[self.hparams.id_label].detach().cpu()
                results.update({self.hparams.id_label: ids, "id": ids})

        return results

    def _step(self, stage, batch, batch_idx, logger):
        (
            x_hat,
            z_parts,
            z_parts_params,
            z_composed,
            loss,
            reconstruction_loss,
            kld_loss,
            kld_per_part,
        ) = self.forward(batch, decode=True, compute_loss=True)

        results = self.make_results_dict(
            stage,
            batch,
            loss,
            reconstruction_loss,
            kld_loss,
            kld_per_part,
            z_parts,
            z_parts_params,
            z_composed,
            x_hat,
        )

        self.log_metrics(stage, results, logger, batch[self.hparams.x_label].shape[0])

        return results



import torch.nn as nn


class SkipConnection(nn.Module):
    def __init__(self, wrapped_module):
        super().__init__()
        self.wrapped_module = wrapped_module

    def forward(self, x):
        return x + self.wrapped_module(x)



import sys
import os

from serotiny.ml_ops import _do_model_op_wrapper


def print_help():
    from textwrap import dedent

    print(
        dedent(
            """
    Usage:
      serotiny COMMAND

    Valid COMMAND values:
      train - train a model
      test - test a model
      predict - use a trained model to output predictions
      config - create a config yaml, given a Python class/function
      dataframe - utils to manipulate .csv dataframes
      image - image operations

    For more info on each command do:")
      serotiny COMMAND --help")
    """
        ).strip()
    )


def main():
    if sys.argv[0].endswith("serotiny"):
        try:
            mode = sys.argv.pop(1)
        except IndexError:
            mode = "help"
    elif sys.argv[0].endswith("serotiny.train"):
        mode = "train"
    elif sys.argv[0].endswith("serotiny.test"):
        mode = "test"
    elif sys.argv[0].endswith("serotiny.predict"):
        mode = "predict"
    else:
        raise NotImplementedError(f"Unknown command: '{sys.argv[0]}")

    if "help" in mode or mode == "-h":
        print_help()
        return

    # hydra modes
    if mode in ("train", "test", "predict"):
        import hydra

        if sys.argv[0].endswith("serotiny"):
            sys.argv[0] += f".{mode}"

        os.environ["HYDRA_FULL_ERROR"] = "1"
        hydra.main(config_path=None, config_name=mode, version_base=None)(
            _do_model_op_wrapper
        )()

    # fire modes
    else:
        from fire import Fire

        import serotiny.cli.config_cli as config_cli
        import serotiny.cli.image_cli as image_cli
        from serotiny.cli.dataframe_cli import DataframeTransformCLI as dataframe_cli

        sys.argv.insert(1, mode)
        cli_dict = {
            "config": config_cli,
            "image": image_cli,
            "dataframe": dataframe_cli,
        }

        if mode in cli_dict:
            Fire(cli_dict)
        else:
            print_help()


if __name__ == "__main__":
    main()



from upath import UPath as Path
from typing import Callable, Optional, Sequence, Type, Union

import aicsimageio
import numpy as np
import torch
from aicsimageio.aics_image import _load_reader
from aicsimageio.writers.ome_tiff_writer import OmeTiffWriter
from ome_zarr.reader import Reader
from ome_zarr.io import parse_url
from omegaconf import ListConfig


def infer_dims(img: aicsimageio.AICSImage):
    """Given a an AICSImage image, infer the corresponding dimensions."""
    dims = dict(img.dims.items())

    if dims.get("T", 1) > 1:
        raise ValueError("Expected image with no T dimensions")
    if (dims["X"] < 1) or (dims["Y"] < 1):
        raise ValueError("Expected image with height and width")

    if dims.get("S", 0) > 1 and dims.get("C", 0) > 1:
        raise ValueError("Can't handle S dimension > 1 AND C dimension > 1")

    zero_dims = {"T": 0}
    order = "YX"
    for dim in list("ZCS"):
        if dims.get(dim, 1) == 1:
            zero_dims[dim] = 0
        else:
            order = dim + order

    if "C" not in order and "S" not in order:
        order = "C" + order
        zero_dims.pop("C")

    return order, zero_dims


def image_loader(
    path,
    select_channels: Optional[Union[Sequence[int], Sequence[str]]] = None,
    dtype: Optional[Union[str, Type[np.number]]] = None,
    transform: Optional[Callable] = None,
    return_channels: bool = False,
    reader: Optional[str] = None,
    unsqueeze_first_dim: bool = False,
    ome_zarr_level: int = 0,
    ome_zarr_image_name: str = "default",
):
    """Load image from path given by `path`. If the given image doesn't have
    channel names, `select_channels` must be a list of integers.

    Parameters
    ----------
    path: str
        Path of the image to load

    select_channels: Optional[Union[Sequence[int], Sequence[str]]] = None
        Channels to be retrieved from the image. If the given image doesn't
        have channel names, `select_channels` must be a list of integers

    dtype: Type[np.number] = np.float32
        Numpy dtype of output image

    transform: Optional[Callable] = None
        Transform to apply before returning the image

    return_channels: bool = False
        Flag to determine whether to return a channel-index map when loading
        the image. This is only useful when channels have names

    unsqueeze_first_dim: bool = False
        Whether to unsqueeze the first dimension.
        Useful when the channel dimension is needed even if there's a single
        channel.

    ome_zarr_level: int = 0
        If reading an ome zarr image, this specifies what level of the pyramid
        to return

    ome_zarr_image_name: str = "default"
        If reading an ome zarr image, this specifies the image name
    """

    if str(path).endswith(".zarr"):
        img = read_ome_zarr(path, ome_zarr_level, ome_zarr_image_name)
    else:
        if reader is not None:
            reader = _load_reader(reader)
        img = aicsimageio.AICSImage(path, reader=reader)
    channel_names = img.channel_names or list(range(img.data.shape[0]))

    if (select_channels is None) or (len(select_channels) == 0):
        select_channels = channel_names

    if not set(select_channels).issubset(channel_names):
        raise KeyError(
            "Some elements of `select_channels` aren't available: "
            f"\tavailable channels: {channel_names}\n"
            f"\tselect_channels: {select_channels}"
        )

    loaded_channels = select_channels
    loaded_channels_idx = [channel_names.index(channel) for channel in loaded_channels]

    order, zero_dims = infer_dims(img)

    first_dim = {order[0]: loaded_channels_idx} if order[0] in ("S", "C") else {}

    for first_dim_key, first_dim_val in first_dim.items():
        if len(first_dim_val) == 1:
            order = order.replace(first_dim_key, "")
            first_dim[first_dim_key] = first_dim_val[0]

    data = img.get_image_data(order, **zero_dims, **first_dim)

    channel_map = {
        channel_name: index for index, channel_name in enumerate(loaded_channels)
    }

    data = data.squeeze()

    if unsqueeze_first_dim:
        data = np.expand_dims(data, axis=0)

    if dtype is not None:
        data = data.astype(dtype)

    if transform:
        if isinstance(transform, (list, tuple, ListConfig)):
            for _transform in transform:
                data = _transform(data)
        else:
            data = transform(data)

    if return_channels:
        return data, channel_map

    return data


def read_ome_zarr(path, level=0, image_name="default"):
    """Function that reads an ome.zarr image and (optionally) returns an
    AICSImage object.

    Note: This is here temporarily, until aicsimageio
    provides an OmeZarrReader class
    """
    path = str(path if image_name is None else Path(path) / image_name)
    reader = Reader(parse_url(path))

    node = next(iter(reader()))
    pps = node.metadata["coordinateTransformations"][0][0]["scale"][-3:]

    return aicsimageio.AICSImage(
        node.data[level].compute(),
        channel_names=node.metadata["name"],
        physical_pixel_sizes=pps,
    )


def tiff_writer(
    img: Union[np.array, torch.Tensor],
    path: Union[str, Path],
    channel_names: Optional[Sequence] = None,
    dim_order: Optional[str] = None,
):
    """Write an image to disk.

    Parameters
    ----------

    img: Union[np.array, torch.Tensor]
        An array/tensor containing the image to be saved

    path: Union[str, Path]
        The path where the image will be saved

    channel_names: Optional[Sequence] = None
        Optional list of labels to assign to each channel
        (in the correct order)
    """

    if dim_order is None:
        if len(img.shape) == 4:
            dim_order = "CZYX"
        elif len(img.shape) == 3:
            dim_order = "CYX"
        else:
            raise ValueError(f"Unexpected image shape {img.shape}")

    if len(dim_order) != len(img.shape):
        raise ValueError(
            f"The dimension spec ({dim_order}) is incompatible "
            f"with the image shape {img.shape}"
        )

    if channel_names is not None:
        channel_names = [channel_names]

    OmeTiffWriter.save(
        data=img, uri=path, channel_names=channel_names, dim_order=dim_order
    )



import logging
from typing import Optional, Sequence, Union

import numpy as np
import torch
import torch.nn as nn
from torch.nn.modules.loss import _Loss as Loss

from serotiny.networks import BasicCNN
from serotiny.networks.vae import ImplicitDecoder
from serotiny.networks.utils import weight_init
import torch.nn.functional as F

from .base_vae import BaseVAE

Array = Union[torch.Tensor, np.ndarray, Sequence[float]]
logger = logging.getLogger("lightning")
logger.propagate = False


class ImplicitDecoderVAE(BaseVAE):
    def __init__(
        self,
        latent_dim: Union[int, Sequence[int]],
        in_channels: int,
        hidden_channels: Sequence[int],
        decoder_hidden_channels: Sequence[int],
        max_pool_layers: Sequence[int],
        input_dims: Sequence[int],
        x_label: str,
        beta: float = 1.0,
        optimizer=torch.optim.Adam,
        id_label: Optional[str] = None,
        non_linearity: Optional[nn.Module] = None,
        decoder_non_linearity: Optional[nn.Module] = None,
        loss_mask_label: Optional[str] = None,
        reconstruction_loss: Loss = nn.MSELoss(reduction="none"),
        reconstruction_reduce: str = "mean",
        prior_mode: str = "isotropic",
        prior_logvar: Optional[Array] = None,
        learn_prior_logvar: bool = False,
        skip_connections: bool = True,
        mode: str = "3d",
        kernel_size: int = 3,
        cache_outputs: Sequence = ("test",),
        temperature: Optional[float] = None,
    ):

        encoder = BasicCNN(
            in_channels=in_channels,
            input_dims=input_dims,
            output_dim=latent_dim * 2,  # because it has to return mu and sigma
            hidden_channels=hidden_channels,
            max_pool_layers=max_pool_layers,
            mode=mode,
            kernel_size=kernel_size,
            non_linearity=non_linearity,
            # final_non_linearity=torch.nn.Sigmoid(),
            skip_connections=skip_connections,
        )
        encoder.apply(weight_init)
        nn.utils.spectral_norm(encoder.output)

        decoder = ImplicitDecoder(
            latent_dims=latent_dim,
            hidden_channels=decoder_hidden_channels,
            non_linearity=non_linearity,
            final_non_linearity=decoder_non_linearity,
            mode=mode,
        )
        decoder.apply(weight_init)

        if decoder_non_linearity is not None:
            decoder = nn.Sequential(decoder, decoder_non_linearity)

        self._current_step = 0
        self._temperature = temperature

        super().__init__(
            encoder=encoder,
            decoder=decoder,
            latent_dim=latent_dim,
            optimizer=optimizer,
            x_label=x_label,
            id_label=id_label,
            loss_mask_label=loss_mask_label,
            beta=beta,
            reconstruction_loss=reconstruction_loss,
            reconstruction_reduce=reconstruction_reduce,
            prior_mode=prior_mode,
            prior_logvar=prior_logvar,
            learn_prior_logvar=learn_prior_logvar,
            cache_outputs=cache_outputs,
        )

    def _current_decoding_dims(self, orig_dims):
        if self._temperature is not None:
            self._current_step = max(self._current_step, self.global_step)

            # TODO: add more flexibility for "scale scheduling"
            self._scale = 1 - 0.5 * np.exp(-self._temperature * self._current_step)
        else:
            self._scale = 1
        return [int(self._scale * dim) for dim in orig_dims]

    def calculate_elbo(self, x, x_hat, mu, logvar, mask=None):
        x_hat = F.interpolate(x_hat, size=x.shape[2:])
        return super().calculate_elbo(x, x_hat, mu, logvar, mask)

    def log_metrics(self, stage, reconstruction_loss, kld_loss, loss, logger):
        self.log(f"{stage} reconstruction loss", reconstruction_loss, logger=logger)
        self.log(f"{stage} kld loss", kld_loss, logger=logger)
        self.log(f"{stage}_loss", loss, logger=logger)
        self.log(f"{stage}_scale", self._scale, logger=logger)
        self.log(f"{stage}_step", self._current_step, logger=logger)

    def decode(self, z_params, **kwargs):
        z = self.sample_z(z_params)
        x_hat = self.decoder_non_linearity(
            self.decoder(
                z, **{k: v for k, v in kwargs.items() if k in self.decoder_args}
            )
        )

        return z, x_hat

    def parse_batch(self, batch):
        x, kwargs = super().parse_batch(batch)

        kwargs["input_dims"] = self._current_decoding_dims(x.shape[2:])
        return x, kwargs



from typing import Optional

from torch import nn


def conv_block(
    in_c: int,
    out_c: int,
    kernel_size: int = 3,
    stride: int = 1,
    padding: int = 0,
    up_conv: bool = False,
    non_linearity: Optional[nn.Module] = None,
    batch_norm: bool = True,
    mode: str = "3d",
):
    """Util function to instantiate a convolutional block.

    Parameters
    ----------
    in_c: int
        number of input channels
    out_c: int
        number of output channels
    kernel_size: Sequence[int] (defaults to (3, 3, 3))
        dimensions of the convolutional kernel to be applied
    stride: Sequence[int] (defaults to (1, 1, 1))
        stride of the convolution
    padding:
        padding value for the convolution (defaults to 0)
    mode:
        Dimensionality of the input data. Can be "2d" or "3d".
    """
    batch_norm_cls = nn.Identity
    if mode == "2d":
        conv = nn.ConvTranspose2d if up_conv else nn.Conv2d
        if batch_norm:
            batch_norm_cls = nn.BatchNorm2d
    elif mode == "3d":
        conv = nn.ConvTranspose3d if up_conv else nn.Conv3d
        if batch_norm:
            batch_norm_cls = nn.BatchNorm3d
    else:
        raise ValueError(f"Mode must be '2d' or '3d'. You passed '{mode}'")

    if non_linearity is None:
        non_linearity = nn.ReLU()

    block = nn.Sequential(
        conv(in_c, out_c, kernel_size=kernel_size, stride=stride, padding=padding),
        non_linearity,
        batch_norm_cls(out_c),
    )

    return block


class ConvBlock(nn.Module):
    def __init__(
        self,
        in_c: int,
        out_c: int,
        kernel_size: int = 3,
        stride: int = 1,
        padding: int = 0,
        up_conv: bool = False,
        non_linearity: Optional[nn.Module] = None,
        skip_connection: bool = False,
        batch_norm: bool = False,
        mode: str = "3d",
    ):
        """Convolutional block class.

        Parameters
        ----------
        in_c: int
            number of input channels
        out_c: int
            number of output channels
        kernel_size: Sequence[int] (defaults to (3, 3, 3))
            dimensions of the convolutional kernel to be applied
        stride: Sequence[int] (defaults to (1, 1, 1))
            stride of the convolution
        padding:
            padding value for the convolution (defaults to 0)
        mode:
            Dimensionality of the input data. Can be "2d" or "3d".
        """

        super().__init__()
        self.skip_connection = skip_connection
        self.block = conv_block(
            in_c=in_c,
            out_c=out_c,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            up_conv=up_conv,
            non_linearity=non_linearity,
            batch_norm=batch_norm,
            mode=mode,
        )

    def forward(self, x):
        res = self.block(x)
        if self.skip_connection and (res.shape[1] == x.shape[1]):
            return res + nn.functional.interpolate(x, res.shape[2:])

        return res



import os
import re
import sys
from typing import Optional

from hydra.plugins.completion_plugin import CompletionPlugin


def strip_serotiny_call(line: str):
    regex = r"^serotiny(?:\s*|\.)(?:train|test|predict)\s*(.*)"
    match = re.match(regex, line)
    if match:
        return match.group(1)
    else:
        raise RuntimeError(f"Error parsing line '{line}'")


class SerotinyCompletion(CompletionPlugin):
    def install(self) -> None:
        script = """hydra_serotiny_bash_completion()
{
    words=($COMP_LINE)
    helper="serotiny.train"
    EXECUTABLE=($(command -v $helper))
    if [ "$HYDRA_COMP_DEBUG" == "1" ]; then
        printf "EXECUTABLE_FIRST='${EXECUTABLE[0]}'\\n"
    fi
    if ! [ -x "${EXECUTABLE[0]}" ]; then
        false
    fi
    if [ $? == 0 ]; then
        choices=$( COMP_POINT=$COMP_POINT COMP_LINE=$COMP_LINE $helper -sc query=serotiny_bash)
        word=${words[$COMP_CWORD]}
        if [ "$HYDRA_COMP_DEBUG" == "1" ]; then
            printf "\\n"
            printf "COMP_LINE='$COMP_LINE'\\n"
            printf "COMP_POINT='$COMP_POINT'\\n"
            printf "Word='$word'\\n"
            printf "Output suggestions:\\n"
            printf "\\t%s\\n" ${choices[@]}
        fi
        COMPREPLY=($( compgen -o nospace -o default -W "$choices" -- "$word" ));
    fi
}
COMP_WORDBREAKS=${COMP_WORDBREAKS//=}
COMP_WORDBREAKS=$COMP_WORDBREAKS complete -o nospace -o default -F hydra_serotiny_bash_completion serotiny train
COMP_WORDBREAKS=$COMP_WORDBREAKS complete -o nospace -o default -F hydra_serotiny_bash_completion serotiny.train
COMP_WORDBREAKS=$COMP_WORDBREAKS complete -o nospace -o default -F hydra_serotiny_bash_completion serotiny test
COMP_WORDBREAKS=$COMP_WORDBREAKS complete -o nospace -o default -F hydra_serotiny_bash_completion serotiny.test
COMP_WORDBREAKS=$COMP_WORDBREAKS complete -o nospace -o default -F hydra_serotiny_bash_completion serotiny predict
COMP_WORDBREAKS=$COMP_WORDBREAKS complete -o nospace -o default -F hydra_serotiny_bash_completion serotiny.predict
"""  # noqa
        print(script)

    def uninstall(self) -> None:
        print("unset hydra_serotiny_bash_completion")

    @staticmethod
    def provides() -> str:
        return "serotiny_bash"

    @staticmethod
    def help(command: str) -> str:
        assert command in ["install", "uninstall"]
        return f'eval "$({{}} -sc {command}=serotiny_bash)"'

    @staticmethod
    def _get_exec() -> str:
        if "train" in sys.argv[0]:
            return "serotiny.train"
        elif "test" in sys.argv[0]:
            return "serotiny.test"
        elif "predict" in sys.argv[0]:
            return "serotiny.test"
        else:
            return None

    def query(self, config_name: Optional[str]) -> None:
        line = os.environ["COMP_LINE"]
        line = strip_serotiny_call(line)

        if "train" in sys.argv[0]:
            config_name = "train"
        elif "test" in sys.argv[0]:
            config_name = "test"
        elif "predict" in sys.argv[0]:
            config_name = "predict"
        else:
            raise RuntimeError(
                "Didn't get a valid config_name. "
                "(Should be 'train', 'test' or 'predict')"
            )

        print(" ".join(self._query(config_name=config_name, line=line)))



import math

import numpy as np
import torch
import torch.nn.functional as F


def split_number(n):
    middle = n / 2
    bottom = math.floor(middle)
    top = math.ceil(middle)
    return (bottom, top)


def pull_to(array, axes: int):
    if not isinstance(array, np.ndarray):
        array = np.array(array)
    missing = axes - len(array.shape)
    pull = np.expand_dims(array, tuple(range(missing)))
    return pull


def expand_to(array, dimensions, pad=None):
    pad = pad or {}
    in_shape = array.shape
    in_dimensions = len(in_shape)
    missing = len(dimensions) - in_dimensions
    pull = np.expand_dims(array, tuple(range(missing)))
    around = [
        split_number(dimensions[dimension] - pull.shape[dimension])
        for dimension in range(in_dimensions)
    ]
    expand = np.pad(pull, around, **pad)
    return expand


def to_tensor(value):
    if isinstance(value, np.ndarray):
        return torch.from_numpy(value)
    return torch.Tensor([value])


def expand_columns(rows, expanded_columns, dimensions, pad=None):
    if not rows:
        return []

    first_row = rows[0]
    all_columns = list(first_row.keys())
    collated = {column: [] for column in all_columns}

    for row in rows:
        for column in all_columns:
            value = row[column]
            if column in expanded_columns:
                value = expand_to(value, dimensions, pad)
            tensor = to_tensor(value)
            collated[column].append(tensor)

    for column in all_columns:
        collated[column] = torch.stack(collated[column])

    return collated


class ExpandTo:
    def __init__(self, dimensions, pad=None):
        self.dimensions = dimensions
        self.pad = pad or {}

    def __call__(self, in_array: np.ndarray):
        return expand_to(in_array, self.dimensions, self.pad)


class ExpandColumns:
    def __init__(self, columns, dimensions, pad=None):
        self.columns = columns
        self.dimensions = dimensions
        self.pad = pad or {}

    def __call__(self, rows):

        return expand_columns(rows, self.columns, self.dimensions, self.pad)


class PadTo:
    def __init__(self, target_dims, mode="constant", value=0):
        self.target_dims = target_dims
        self.mode = mode
        self.value = value

    def __call__(self, img):
        if not isinstance(img, torch.Tensor):
            img = torch.tensor(img)

        pad = []
        for i, dim in enumerate(self.target_dims):
            pad_dim = (dim - img.shape[i + 1]) / 2
            pad_dim = max(0, pad_dim)

            # when 2 * pad_dim is even, this doesn't change the result.
            # when 2 * pad_dim is odd, this makes padding amount one pixel/voxel
            # bigger on one side
            pad.extend((math.floor(pad_dim), math.ceil(pad_dim)))

        # pytorch pad function expects padding amount in reverse order
        pad = pad[::-1]

        return F.pad(img, pad, mode=self.mode, value=self.value)



import logging
from typing import Optional, Sequence, Union

import numpy as np
import torch
import torch.nn as nn
from torch.nn.modules.loss import _Loss as Loss

from serotiny.networks import BasicCNN
from serotiny.networks.vae import ImageDecoderBasicCNN
from serotiny.networks.utils import weight_init
import serotiny

from .base_vae import BaseVAE
from .priors import Prior

Array = Union[torch.Tensor, np.ndarray, Sequence[float]]
logger = logging.getLogger("lightning")
logger.propagate = False
from omegaconf import DictConfig


class ImageVAE(BaseVAE):
    def __init__(
        self,
        latent_dim: Union[int, Sequence[int]],
        in_channels: int,
        hidden_channels: Sequence[int],
        max_pool_layers: Sequence[int],
        input_dims: Sequence[int],
        x_label: str,
        optimizer: torch.optim.Optimizer = torch.optim.Adam,
        beta: float = 1.0,
        id_label: Optional[str] = None,
        non_linearity: Optional[nn.Module] = None,
        decoder_non_linearity: Optional[nn.Module] = None,
        loss_mask_label: Optional[str] = None,
        reconstruction_loss: Loss = nn.MSELoss(reduction="none"),
        skip_connections: bool = True,
        batch_norm: bool = True,
        mode: str = "3d",
        prior: Optional[Sequence[Prior]] = None,
        kernel_size: int = 3,
        cache_outputs: Sequence = ("test",),
        encoder_clamp: Optional[int] = 6,
        # final_non_linearity: Optional[nn.Module] = None,
    ):
        if isinstance(
            prior[x_label], serotiny.models.vae.priors.IsotropicGaussianPrior
        ):
            encoder_latent_dims = 2 * latent_dim
        else:
            encoder_latent_dims = latent_dim

        encoder = BasicCNN(
            in_channels=in_channels,
            input_dims=input_dims,
            output_dim=encoder_latent_dims,  # because it has to return mu and sigma
            hidden_channels=hidden_channels,
            max_pool_layers=max_pool_layers,
            mode=mode,
            kernel_size=kernel_size,
            non_linearity=non_linearity,
            skip_connections=skip_connections,
            batch_norm=batch_norm,
            encoder_clamp=encoder_clamp,
        )
        encoder.apply(weight_init)

        dummy_out, intermediate_sizes = encoder.conv_forward(
            torch.zeros(1, in_channels, *input_dims), return_sizes=True
        )

        compressed_img_shape = dummy_out.shape[2:]

        intermediate_sizes = [input_dims] + intermediate_sizes[:-1]
        intermediate_sizes = intermediate_sizes[::-1]

        decoder = ImageDecoderBasicCNN(
            compressed_img_shape=compressed_img_shape,
            hidden_channels=list(reversed(hidden_channels)),
            intermediate_sizes=intermediate_sizes,
            latent_dim=latent_dim,
            output_dims=tuple(input_dims),
            output_channels=in_channels,
            mode=mode,
            non_linearity=non_linearity,
            skip_connections=skip_connections,
            batch_norm=batch_norm,
            kernel_size=kernel_size,
        )
        decoder.apply(weight_init)
        nn.utils.spectral_norm(decoder.linear_decompress)

        if decoder_non_linearity is not None:
            decoder = nn.Sequential(decoder, decoder_non_linearity)

        if not isinstance(encoder, (dict, DictConfig)):
            assert x_label is not None
            encoder = {x_label: encoder}

        if not isinstance(decoder, (dict, DictConfig)):
            assert x_label is not None
            decoder = {x_label: decoder}

        if not isinstance(reconstruction_loss, (dict, DictConfig)):
            assert x_label is not None
            reconstruction_loss = {x_label: reconstruction_loss}

        if not isinstance(prior, (dict, DictConfig)):
            assert x_label is not None
            prior = {x_label: prior}

        super().__init__(
            encoder=encoder,
            decoder=decoder,
            latent_dim=latent_dim,
            optimizer=optimizer,
            x_label=x_label,
            id_label=id_label,
            beta=beta,
            reconstruction_loss=reconstruction_loss,
            cache_outputs=cache_outputs,
            optimizier=optimizer,
            prior=prior,
        )



import numpy as np

import torch.nn as nn
import torch


class ImageDecoder(nn.Module):
    def __init__(
        self,
        encoder,
        deconv_block,
        input_dims,
        in_channels,
        latent_dim,
    ):
        super().__init__()
        inp = torch.zeros(1, in_channels, *input_dims)
        compressed_img_shape = encoder(inp).shape
        self.compressed_img_shape = compressed_img_shape[2:]

        deconv_layers = []

        if isinstance(encoder[0][0], torch.nn.Conv3d):
            mode = "3d"
        elif isinstance(encoder[0][0], torch.nn.Conv2d):
            mode = "2d"

        for index in range(len(encoder) - 1, -1, -1):
            conv = encoder[index][0]
            if index == len(encoder) - 1:
                self.first_out_channels = conv.out_channels
                compressed_img_size = (
                    np.prod(self.compressed_img_shape) * self.first_out_channels
                )
            if index == 0:
                out_channels = in_channels
            else:
                out_channels = conv.out_channels

            if mode == "3d":
                deconv_layers.append(
                    torch.nn.LazyConvTranspose3d(
                        out_channels=out_channels,
                        kernel_size=conv.kernel_size,
                        stride=conv.stride,
                    )
                )

            else:
                deconv_layers.append(
                    torch.nn.LazyConvTranspose2d(
                        out_channels=out_channels,
                        kernel_size=conv.kernel_size,
                        stride=conv.stride,
                    )
                )
            if index != 0:
                deconv_layers.append(deconv_block)

        deconv = nn.Sequential(*deconv_layers)
        self.deconv = deconv
        self.linear_decompress = nn.Linear(latent_dim, compressed_img_size)

    def forward(self, z):
        z = self.linear_decompress(z)
        z = z.view(
            z.shape[0],  # batch size
            self.first_out_channels,
            *self.compressed_img_shape
        )
        z = self.deconv(z)
        z = z.clamp(max=50)
        return z



import torch.nn as nn
import torch.nn.init as init


def weight_init(m):
    """
    Usage:
        model = Model()
        model.apply(weight_init)
    """
    if isinstance(m, nn.Conv1d):
        init.normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.Conv2d):
        init.xavier_normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.Conv3d):
        init.xavier_normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.ConvTranspose1d):
        init.normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.ConvTranspose2d):
        init.xavier_normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.ConvTranspose3d):
        init.xavier_normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.BatchNorm1d):
        init.normal_(m.weight.data, mean=1, std=0.02)
        init.constant_(m.bias.data, 0)
    elif isinstance(m, nn.BatchNorm2d):
        init.normal_(m.weight.data, mean=1, std=0.02)
        init.constant_(m.bias.data, 0)
    elif isinstance(m, nn.BatchNorm3d):
        init.normal_(m.weight.data, mean=1, std=0.02)
        init.constant_(m.bias.data, 0)
    elif isinstance(m, nn.Linear):
        init.xavier_normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.LSTM):
        for param in m.parameters():
            if len(param.shape) >= 2:
                init.orthogonal_(param.data)
            else:
                init.normal_(param.data)
    elif isinstance(m, nn.LSTMCell):
        for param in m.parameters():
            if len(param.shape) >= 2:
                init.orthogonal_(param.data)
            else:
                init.normal_(param.data)
    elif isinstance(m, nn.GRU):
        for param in m.parameters():
            if len(param.shape) >= 2:
                init.orthogonal_(param.data)
            else:
                init.normal_(param.data)
    elif isinstance(m, nn.GRUCell):
        for param in m.parameters():
            if len(param.shape) >= 2:
                init.orthogonal_(param.data)
            else:
                init.normal_(param.data)


if __name__ == "__main__":
    pass



import torch.nn as nn

_full_slice = slice(None, None, None)


class Prior(nn.Module):
    def __init__(self, dimensionality: int, **kwargs):
        super().__init__()
        self.dimensionality = dimensionality

    def __len__(self):
        return self.dimensionality

    def kl_divergence(self, z):
        raise NotImplementedError

    def sample(self, z):
        raise NotImplementedError



import torch.nn.functional as F
from torch.nn.modules.loss import _Loss as Loss


class CosineLoss(Loss):
    def __init__(self, reduction="mean"):
        super(CosineLoss, self).__init__(None, None, reduction)

    def forward(self, input, target):

        # sum per input-element log loss
        loss = 1 - F.cosine_similarity(input, target)

        # reduce across batch dimension
        if self.reduction == "none":
            return loss
        elif self.reduction == "sum":
            return loss.sum()
        elif self.reduction == "mean":
            return loss.mean()
        else:
            raise NotImplementedError(f"Unavailable reduction type: {self.reduction}")



import torch.nn as nn


class MultiplyConstant(nn.Module):
    def __init__(self, constant):
        super().__init__()
        self.constant = constant

    def forward(self, x):
        return self.constant * x



import torch
from torch import nn
from typing import Optional
from serotiny.networks.utils import weight_init


def _make_block(input_dim, output_dim):
    return nn.Sequential(
        nn.Linear(input_dim, output_dim),
        nn.BatchNorm1d(output_dim),
        nn.ReLU(),
    )


class MLP(nn.Module):
    def __init__(
        self,
        *dims,
        hidden_layers=[256],
        scale_output: Optional[int] = 1,
    ):
        super().__init__()

        assert len(dims) >= 2

        self.input_dims = dims[:-1]
        self.output_dim = dims[-1] * scale_output

        self.hidden_layers = hidden_layers

        net = [_make_block(sum(self.input_dims), hidden_layers[0])]

        net += [
            _make_block(input_dim, output_dim)
            for (input_dim, output_dim) in zip(hidden_layers[0:], hidden_layers[1:])
        ]

        net += [nn.Linear(hidden_layers[-1], self.output_dim)]

        self.net = nn.Sequential(*net)
        self.net.apply(weight_init)

    def forward(self, x1, condition=None):
        if condition is not None:
            if len(condition.shape) == 1:
                condition = condition.unsqueeze(dim=1)
            x1 = torch.cat([x1, condition], dim=1)
        return self.net(x1)



import torch
from .abstract_prior import Prior


class IdentityPrior(Prior):
    """Prior class that doesn't contribute to KL loss.

    Effectively a Dirac delta distribution given z.
    """

    def forward(self, z, mode="kl", **kwargs):
        if mode == "kl":
            return torch.tensor(0).float()
        return z
